function [F_classify_opt,F_regions_opt]=local_lda(xtrain,ytrain,n_stages,max_iterations,n_inits,w)
% This code is based on the paper: J. Wang, V. Saligrama, Local Supervised
% Learning through Space Partitioning, in NIPS, pp. 91-99, 2012.

% Inputs:
% xtrain: d x n training data, where each column is a training example
% ytrain: 1 x n training labels
% n_stages: number of local linear classifiers
% max_iterations: maximum number of alternating minimization steps per initialization (default is 25)
% n_inits: number of random initializations (default is 15)
% w: weights on training points (default is uniform weight)

% Outputs:
% F_classify_opt: local classification structure
% F_regions_opt: local partitioning structure

if nargin<4
    max_iterations=25;
end
if nargin<5
    n_inits=15;
end
if nargin<6
    w=ones(1,size(xtrain,2))/size(xtrain,2);
end



min_error=inf;
F_classify_opt=struct('labels',{},'mu',{},'pi',{},'sigma_inv',{});
F_regions_opt=struct('labels',{},'mu',{},'pi',{},'sigma_inv',{});
for initialization=1:n_inits
    
%{    
    regions=zeros(size(ytrain));
    for i=1:(n_stages-1)
        cut_dim=randi(size(xtrain,1));
    
        thresh_vec=xtrain(cut_dim,:);
        thresh_vec(regions~=0)=[];
        thresh=sort(thresh_vec);
        thresh=thresh(ceil(length(ytrain)/n_stages));
        temp_labels=zeros(size(regions));
        temp_labels(xtrain(cut_dim,:)<thresh)=i;
        regions=regions+(regions==0).*temp_labels;
    end
    regions(regions==0)=n_stages;
%}    
    regions=randi(n_stages,size(ytrain));
    
    
    
    F_classify=struct('labels',{},'mu',{},'pi',{},'sigma_inv',{});
    F_regions=struct('labels',{},'mu',{},'pi',{},'sigma_inv',{});
    
    it=1;
    
    yhat_all=zeros(n_stages,length(ytrain));
    yhat=zeros(size(ytrain));
    
    
    for i=1:n_stages
        F_classify(i)=weighted_multiclass_lda(xtrain(:,regions==i),ytrain(regions==i),w(regions==i));
        yhat_all(i,:)=eval_multiclass_lda(F_classify(i),xtrain);
    end
    
    
    
    while it<=max_iterations
%         disp(['Training Iteration ' num2str(it)]);
        old_region_labels=regions;
        old_labels=yhat;
        
        label_matches=(yhat_all==repmat(ytrain,n_stages,1));
        regions=n_stages*ones(size(ytrain));
        for i=((n_stages-1):-1:1)
            active_points=(regions>=i);
            matches_stay=label_matches(i,:);
            if i<(n_stages-1)
                matches_go=eval_local_lda(F_classify(i+1:end),F_regions(i+1:end),xtrain)==ytrain;
            else
                matches_go=label_matches(end,:);
            end
            weights=w.*(matches_stay~=matches_go).*active_points;
            temp_labels=ones(size(ytrain));
            temp_labels(matches_stay)=-1;
            F_regions(i)=weighted_multiclass_lda(xtrain,temp_labels,weights);
            temp=eval_multiclass_lda(F_regions(i),xtrain);
            region_est=n_stages*ones(size(ytrain));
            region_est(temp==-1)=i;
            regions=min(regions,region_est);
        end
        
        for i=1:n_stages
            F_classify(i)=weighted_multiclass_lda(xtrain(:,regions==i),ytrain(regions==i),w(regions==i));
            yhat_all(i,:)=eval_multiclass_lda(F_classify(i),xtrain);
        end
        
        
        [yhat regions]=eval_local_lda(F_classify,F_regions,xtrain);
        
%         disp(['    ' num2str(length(unique(regions))) ' regions remaining, training error:' num2str(mean(yhat~=ytrain)*100) ' %']);        
        
        if (sum(regions~=old_region_labels)+sum(yhat~=old_labels))==0
%             disp('No change in region or classification labels, terminating iteration early');
            break;
        elseif sum(yhat~=ytrain)==0 && sum(old_labels~=ytrain)==0
%             disp('0% training error, terminating early');
            break;
        end
        
        it=it+1;
    end
    
    disp(['Initialization: ' num2str(initialization) ', training error: ' num2str(mean(yhat~=ytrain))]);
    if mean(yhat~=ytrain)<min_error
        min_error=mean(yhat~=ytrain);
        F_classify_opt=F_classify;
        F_regions_opt=F_regions;
    end
    if min_error==0
        break;
    end
end

% [yhat regions]=eval_local_lda(F_classify_opt,F_regions_opt,xtrain);
% disp(['Local LDA trained for ' num2str(length(unique(regions))) ' local regions, training error: ' num2str(mean(yhat~=ytrain)*100) ' %']);




function F=weighted_multiclass_lda(xtrain,ytrain,w)

if nargin<3
    w=ones(1,size(xtrain,2));
end


F=struct('labels',{},'mu',{},'pi',{},'sigma_inv',{});


[labels,~,y]=unique(ytrain);

F(1).labels=labels;
F(1).sigma_inv=eye(size(xtrain,1));
F(1).pi=zeros(length(labels),1);
F(1).mu=zeros(size(xtrain,1),length(labels));

if sum(w)==0
    return;
elseif length(labels)==1
    return;
else
    w=w./sum(w);
end
temp_pi=zeros(length(labels),1);
temp_mu=zeros(size(xtrain,1),length(labels));
temp_sigma=zeros(size(xtrain,1));
for i=1:length(labels)
    temp_pi(i)=sum(w(y==i));
    if temp_pi(i)~=0
        temp_mu(:,i)=sum(repmat(w(y==i)./sum(w(y==i)),size(xtrain,1),1).*xtrain(:,y==i),2);
        temp_matrix=((xtrain(:,y==i)-repmat(temp_mu(:,i),1,sum(y==i))).*repmat(w(y==i),size(xtrain,1),1))*(xtrain(:,y==i)-repmat(temp_mu(:,i),1,sum(y==i)))'/(1-sum(w(y==i)));
        
        if sum(isnan(temp_matrix(:)))==0 && sum(isinf(temp_matrix(:)))==0
            temp_sigma=temp_sigma+temp_matrix;
        end
    end
end
F(1).sigma_inv=pinv(temp_sigma);
F(1).pi=temp_pi;
F(1).mu=temp_mu;



function [yhat deltas]=eval_multiclass_lda(F,x)

if isempty(F(1).labels)
    yhat=zeros(1,size(x,2));
    deltas=zeros(1,size(x,2));
    return;
end


if size(.5*diag(F(1).mu'*F(1).sigma_inv*F(1).mu),1)==0
    disp('trouble');
end


offsets=.5*diag(F(1).mu'*F(1).sigma_inv*F(1).mu)-log(F(1).pi);

deltas=(x'*F(1).sigma_inv*F(1).mu)'-repmat(offsets,1,size(x,2));

[~,ind]=max(deltas,[],1);
labels=F(1).labels;
yhat=labels(ind);


